@@ -25,12 +25,12 @@
  * @brief SFTP handling functions
  *
  * SFTP commands are channeled by the ssh sftp subsystem. Every packet is
- * sent/read using a SFTP_PACKET type structure. Related to these packets,
+ * sent/read using a sftp_packet type structure. Related to these packets,
  * most of the server answers are messages having an ID and a message
- * specific part. It is described by SFTP_MESSAGE when reading a message,
+ * specific part. It is described by sftp_message when reading a message,
  * the sftp system puts it into the queue, so the process having asked for
  * it can fetch it, while continuing to read for other messages (it is
- * inspecified in which order messages may be sent back to the client
+ * unspecified in which order messages may be sent back to the client
  *
  * @defgroup ssh_sftp SFTP Functions
  * @{
@@ -38,126 +38,156 @@
 
 #ifndef SFTP_H
 #define SFTP_H
-#include <libssh/libssh.h>
+
+#include <sys/types.h>
+
+#include "libssh.h"
+
 #ifdef __cplusplus
 extern "C" {
 #endif
 
-#ifdef __GNUC__
-#define SFTP_DEPRECATED __attribute__ ((deprecated))
-#else
-#define SFTP_DEPRECATED
-#endif
-
 #ifdef _WIN32
 #ifndef uid_t
-  typedef long uid_t;
+  typedef uint32_t uid_t;
 #endif /* uid_t */
 #ifndef gid_t
-  typedef long gid_t;
+  typedef uint32_t gid_t;
 #endif /* gid_t */
+#ifdef _MSC_VER
+#ifndef ssize_t
+  typedef _W64 SSIZE_T ssize_t;
+#endif /* ssize_t */
+#endif /* _MSC_VER */
 #endif /* _WIN32 */
 
-typedef struct sftp_session_struct {
-    SSH_SESSION *session;
-    CHANNEL *channel;
+typedef struct sftp_attributes_struct* sftp_attributes;
+typedef struct sftp_client_message_struct* sftp_client_message;
+typedef struct sftp_dir_struct* sftp_dir;
+typedef struct sftp_ext_struct *sftp_ext;
+typedef struct sftp_file_struct* sftp_file;
+typedef struct sftp_message_struct* sftp_message;
+typedef struct sftp_packet_struct* sftp_packet;
+typedef struct sftp_request_queue_struct* sftp_request_queue;
+typedef struct sftp_session_struct* sftp_session;
+typedef struct sftp_status_message_struct* sftp_status_message;
+typedef struct sftp_statvfs_struct* sftp_statvfs_t;
+
+struct sftp_session_struct {
+    ssh_session session;
+    ssh_channel channel;
     int server_version;
     int client_version;
     int version;
-    struct request_queue *queue;
-    u32 id_counter;
+    sftp_request_queue queue;
+    uint32_t id_counter;
     int errnum;
     void **handles;
-} SFTP_SESSION ;
+    sftp_ext ext;
+};
 
-typedef struct {
-    SFTP_SESSION *sftp;
-    u8 type;
-    BUFFER *payload;
-} SFTP_PACKET;
+struct sftp_packet_struct {
+    sftp_session sftp;
+    uint8_t type;
+    ssh_buffer payload;
+};
 
 /* file handler */
-typedef struct sftp_file{
-    SFTP_SESSION *sftp;
+struct sftp_file_struct {
+    sftp_session sftp;
     char *name;
-    u64 offset;
-    STRING *handle;
+    uint64_t offset;
+    ssh_string handle;
     int eof;
     int nonblocking;
-}  SFTP_FILE ;
+};
 
-typedef struct sftp_dir {
-    SFTP_SESSION *sftp;
+struct sftp_dir_struct {
+    sftp_session sftp;
     char *name;
-    STRING *handle; /* handle to directory */
-    BUFFER *buffer; /* contains raw attributes from server which haven't been parsed */
-    u32 count; /* counts the number of following attributes structures into buffer */
+    ssh_string handle; /* handle to directory */
+    ssh_buffer buffer; /* contains raw attributes from server which haven't been parsed */
+    uint32_t count; /* counts the number of following attributes structures into buffer */
     int eof; /* end of directory listing */
-} SFTP_DIR;
+};
 
-typedef struct {
-    SFTP_SESSION *sftp;
-    u8 packet_type;
-    BUFFER *payload;
-    u32 id;
-} SFTP_MESSAGE;
+struct sftp_message_struct {
+    sftp_session sftp;
+    uint8_t packet_type;
+    ssh_buffer payload;
+    uint32_t id;
+};
 
 /* this is a bunch of all data that could be into a message */
-typedef struct sftp_client_message{
-    SFTP_SESSION *sftp;
-    u8 type;
-    u32 id;
+struct sftp_client_message_struct {
+    sftp_session sftp;
+    uint8_t type;
+    uint32_t id;
     char *filename; /* can be "path" */
-    u32 flags;
-    struct sftp_attributes *attr;
-    STRING *handle;
-    u64 offset;
-    u32 len;
+    uint32_t flags;
+    sftp_attributes attr;
+    ssh_string handle;
+    uint64_t offset;
+    uint32_t len;
     int attr_num;
-    BUFFER *attrbuf; /* used by sftp_reply_attrs */
-    STRING *data; /* can be newpath of rename() */
-} SFTP_CLIENT_MESSAGE;
+    ssh_buffer attrbuf; /* used by sftp_reply_attrs */
+    ssh_string data; /* can be newpath of rename() */
+};
 
-typedef struct request_queue{
-    struct request_queue *next;
-    SFTP_MESSAGE *message;
-} REQUEST_QUEUE;
+struct sftp_request_queue_struct {
+    sftp_request_queue next;
+    sftp_message message;
+};
 
 /* SSH_FXP_MESSAGE described into .7 page 26 */
-typedef struct {
-    u32 id;
-    u32 status;
-    STRING *error;
-    STRING *lang;
+struct sftp_status_message_struct {
+		uint32_t id;
+		uint32_t status;
+    ssh_string error;
+    ssh_string lang;
     char *errormsg;
     char *langmsg;
-} STATUS_MESSAGE;
+};
 
 /* don't worry much of these aren't really used */
-typedef struct sftp_attributes{
+struct sftp_attributes_struct {
     char *name;
     char *longname; /* some weird stuff */
-    u32 flags;
-    u8 type;
-    u64 size;
-    u32 uid;
-    u32 gid;
+    uint32_t flags;
+    uint8_t type;
+    uint64_t size;
+    uint32_t uid;
+    uint32_t gid;
     char *owner;
     char *group;
-    u32 permissions;
-    u64 atime64;
-    u32 atime;
-    u32 atime_nseconds;
-    u64 createtime;
-    u32 createtime_nseconds;
-    u64 mtime64;
-    u32 mtime;
-    u32 mtime_nseconds;
-    STRING *acl;
-    u32 extended_count;
-    STRING *extended_type;
-    STRING *extended_data;
-} SFTP_ATTRIBUTES;
+    uint32_t permissions;
+    uint64_t atime64;
+    uint32_t atime;
+    uint32_t atime_nseconds;
+    uint64_t createtime;
+    uint32_t createtime_nseconds;
+    uint64_t mtime64;
+    uint32_t mtime;
+    uint32_t mtime_nseconds;
+    ssh_string acl;
+    uint32_t extended_count;
+    ssh_string extended_type;
+    ssh_string extended_data;
+};
+
+struct sftp_statvfs_struct {
+  uint64_t f_bsize; /* file system block size */
+  uint64_t f_frsize; /* fundamental fs block size */
+  uint64_t f_blocks; /* number of blocks (unit f_frsize) */
+  uint64_t f_bfree; /* free blocks in file system */
+  uint64_t f_bavail; /* free blocks for non-root */
+  uint64_t f_files; /* total file inodes */
+  uint64_t f_ffree; /* free file inodes */
+  uint64_t f_favail; /* free file inodes for to non-root */
+  uint64_t f_fsid; /* file system id */
+  uint64_t f_flag; /* bit mask of f_flag values */
+  uint64_t f_namemax; /* maximum filename length */
+};
 
 #define LIBSFTP_VERSION 3
 
@@ -168,14 +198,14 @@ typedef struct sftp_attributes{
  *
  * @return              A new sftp session or NULL on error.
  */
-SFTP_SESSION *sftp_new(SSH_SESSION *session);
+LIBSSH_API sftp_session sftp_new(ssh_session session);
 
 /**
  * @brief Close and deallocate a sftp session.
  *
  * @param sftp          The sftp session handle to free.
  */
-void sftp_free(SFTP_SESSION *sftp);
+LIBSSH_API void sftp_free(sftp_session sftp);
 
 /**
  * @brief Initialize the sftp session with the server.
@@ -184,7 +214,7 @@ void sftp_free(SFTP_SESSION *sftp);
  *
  * @return              0 on success, < 0 on error with ssh error set.
  */
-int sftp_init(SFTP_SESSION *sftp);
+LIBSSH_API int sftp_init(sftp_session sftp);
 
 /**
  * @brief Get the last sftp error.
@@ -196,11 +226,65 @@ int sftp_init(SFTP_SESSION *sftp);
  * @return              The saved error (see server responses), < 0 if an error
  *                      in the function occured.
  */
-int sftp_get_error(SFTP_SESSION *sftp);
+LIBSSH_API int sftp_get_error(sftp_session sftp);
+
+/**
+ * @brief Get the count of extensions provided by the server.
+ *
+ * @param  sftp         The sftp session to use.
+ *
+ * @return The count of extensions provided by the server, 0 on error or
+ *         not available.
+ */
+LIBSSH_API unsigned int sftp_extensions_get_count(sftp_session sftp);
+
+/**
+ * @brief Get the name of the extension provided by the server.
+ *
+ * @param  sftp         The sftp session to use.
+ *
+ * @param  indexn        The index number of the extension name you want.
+ *
+ * @return              The name of the extension.
+ */
+LIBSSH_API const char *sftp_extensions_get_name(sftp_session sftp, unsigned int indexn);
+
+/**
+ * @brief Get the data of the extension provided by the server.
+ *
+ * This is normally the version number of the extension.
+ *
+ * @param  sftp         The sftp session to use.
+ *
+ * @param  indexn        The index number of the extension data you want.
+ *
+ * @return              The data of the extension.
+ */
+LIBSSH_API const char *sftp_extensions_get_data(sftp_session sftp, unsigned int indexn);
+
+/**
+ * @brief Check if the given extension is supported.
+ *
+ * @param  sftp         The sftp session to use.
+ *
+ * @param  name         The name of the extension.
+ *
+ * @param  data         The data of the extension.
+ *
+ * @return 1 if supported, 0 if not.
+ *
+ * Example:
+ *
+ * @code
+ * sftp_extension_supported(sftp, "statvfs@openssh.com", "2");
+ * @endcode
+ */
+LIBSSH_API int sftp_extension_supported(sftp_session sftp, const char *name,
+    const char *data);
 
 /**
  * @brief Open a directory used to obtain directory entries.
- *
+ 
  * @param session       The sftp session handle to open the directory.
  * @param path          The path of the directory to open.
  *
@@ -210,7 +294,7 @@ int sftp_get_error(SFTP_SESSION *sftp);
  * @see                 sftp_readdir
  * @see                 sftp_closedir
  */
-SFTP_DIR *sftp_opendir(SFTP_SESSION *session, const char *path);
+LIBSSH_API sftp_dir sftp_opendir(sftp_session session, const char *path);
 
 /**
  * @brief Get a single file attributes structure of a directory.
@@ -225,7 +309,7 @@ SFTP_DIR *sftp_opendir(SFTP_SESSION *session, const char *path);
  * @see                sftp_attribute_free()
  * @see                sftp_closedir()
  */
-SFTP_ATTRIBUTES *sftp_readdir(SFTP_SESSION *session, SFTP_DIR *dir);
+LIBSSH_API sftp_attributes sftp_readdir(sftp_session session, sftp_dir dir);
 
 /**
  * @brief Tell if the directory has reached EOF (End Of File).
@@ -236,7 +320,7 @@ SFTP_ATTRIBUTES *sftp_readdir(SFTP_SESSION *session, SFTP_DIR *dir);
  *
  * @see                 sftp_readdir()
  */
-int sftp_dir_eof(SFTP_DIR *dir);
+LIBSSH_API int sftp_dir_eof(sftp_dir dir);
 
 /**
  * @brief Get information about a file or directory.
@@ -248,7 +332,7 @@ int sftp_dir_eof(SFTP_DIR *dir);
  * @return              The sftp attributes structure of the file or directory,
  *                      NULL on error with ssh and sftp error set.
  */
-SFTP_ATTRIBUTES *sftp_stat(SFTP_SESSION *session, const char *path);
+LIBSSH_API sftp_attributes sftp_stat(sftp_session session, const char *path);
 
 /**
  * @brief Get information about a file or directory.
@@ -263,7 +347,7 @@ SFTP_ATTRIBUTES *sftp_stat(SFTP_SESSION *session, const char *path);
  * @return              The sftp attributes structure of the file or directory,
  *                      NULL on error with ssh and sftp error set.
  */
-SFTP_ATTRIBUTES *sftp_lstat(SFTP_SESSION *session, const char *path);
+LIBSSH_API sftp_attributes sftp_lstat(sftp_session session, const char *path);
 
 /**
  * @brief Get information about a file or directory from a file handle.
@@ -273,14 +357,14 @@ SFTP_ATTRIBUTES *sftp_lstat(SFTP_SESSION *session, const char *path);
  * @return              The sftp attributes structure of the file or directory,
  *                      NULL on error with ssh and sftp error set.
  */
-SFTP_ATTRIBUTES *sftp_fstat(SFTP_FILE *file);
+LIBSSH_API sftp_attributes sftp_fstat(sftp_file file);
 
 /**
  * @brief Free a sftp attribute structure.
  *
  * @param file          The sftp attribute structure to free.
  */
-void sftp_attributes_free(SFTP_ATTRIBUTES *file);
+LIBSSH_API void sftp_attributes_free(sftp_attributes file);
 
 /**
  * @brief Close a directory handle opened by sftp_opendir().
@@ -289,12 +373,7 @@ void sftp_attributes_free(SFTP_ATTRIBUTES *file);
  *
  * @return              Returns SSH_NO_ERROR or SSH_ERROR if an error occured.
  */
-int sftp_closedir(SFTP_DIR *dir);
-
-/**
- * @deprecated          Use sftp_closedir() instead.
- */
-int sftp_dir_close(SFTP_DIR *dir) SFTP_DEPRECATED;
+LIBSSH_API int sftp_closedir(sftp_dir dir);
 
 /**
  * @brief Close an open file handle.
@@ -305,12 +384,7 @@ int sftp_dir_close(SFTP_DIR *dir) SFTP_DEPRECATED;
  *
  * @see                 sftp_open()
  */
-int sftp_close(SFTP_FILE *file);
-
-/**
- * @deprecated          Use sftp_close() instead.
- */
-int sftp_file_close(SFTP_FILE *file) SFTP_DEPRECATED;
+LIBSSH_API int sftp_close(sftp_file file);
 
 /**
  * @brief Open a file on the server.
@@ -319,7 +393,7 @@ int sftp_file_close(SFTP_FILE *file) SFTP_DEPRECATED;
  *
  * @param file          The file to be opened.
  *
- * @param access        Is one of O_RDONLY, O_WRONLY or O_RDWR which request
+ * @param accesstype        Is one of O_RDONLY, O_WRONLY or O_RDWR which request
  *                      opening  the  file  read-only,write-only or read/write.
  *                      Acesss may also be bitwise-or'd with one or  more of
  *                      the following:
@@ -338,12 +412,12 @@ int sftp_file_close(SFTP_FILE *file) SFTP_DEPRECATED;
  * @return              A sftp file handle, NULL on error with ssh and sftp
  *                      error set.
  */
-SFTP_FILE *sftp_open(SFTP_SESSION *session, const char *file, int flags,
+LIBSSH_API sftp_file sftp_open(sftp_session session, const char *file, int accesstype,
     mode_t mode);
 
-void sftp_file_set_nonblocking(SFTP_FILE *handle);
+LIBSSH_API void sftp_file_set_nonblocking(sftp_file handle);
 
-void sftp_file_set_blocking(SFTP_FILE *handle);
+LIBSSH_API void sftp_file_set_blocking(sftp_file handle);
 
 /**
  * @brief Read from a file using an opened sftp file handle.
@@ -357,7 +431,7 @@ void sftp_file_set_blocking(SFTP_FILE *handle);
  * @return              Number of bytes written, < 0 on error with ssh and sftp
  *                      error set.
  */
-ssize_t sftp_read(SFTP_FILE *file, void *buf, size_t count);
+LIBSSH_API ssize_t sftp_read(sftp_file file, void *buf, size_t count);
 
 /**
  * @brief Start an asynchronous read from a file using an opened sftp file handle.
@@ -390,7 +464,7 @@ ssize_t sftp_read(SFTP_FILE *file, void *buf, size_t count);
  * @see                 sftp_async_read()
  * @see                 sftp_open()
  */
-int sftp_async_read_begin(SFTP_FILE *file, u32 len);
+LIBSSH_API int sftp_async_read_begin(sftp_file file, uint32_t len);
 
 /**
  * @brief Wait for an asynchronous read to complete and save the data.
@@ -415,7 +489,7 @@ int sftp_async_read_begin(SFTP_FILE *file, u32 len);
  *
  * @see sftp_async_read_begin()
  */
-int sftp_async_read(SFTP_FILE *file, void *data, u32 len, u32 id);
+LIBSSH_API int sftp_async_read(sftp_file file, void *data, uint32_t len, uint32_t id);
 
 /**
  * @brief Write to a file using an opened sftp file handle.
@@ -433,7 +507,7 @@ int sftp_async_read(SFTP_FILE *file, void *data, u32 len, u32 id);
  * @see                 sftp_read()
  * @see                 sftp_close()
  */
-ssize_t sftp_write(SFTP_FILE *file, const void *buf, size_t count);
+LIBSSH_API ssize_t sftp_write(sftp_file file, const void *buf, size_t count);
 
 /**
  * @brief Seek to a specific location in a file.
@@ -444,7 +518,7 @@ ssize_t sftp_write(SFTP_FILE *file, const void *buf, size_t count);
  *
  * @return             0 on success, < 0 on error.
  */
-int sftp_seek(SFTP_FILE *file, u32 new_offset);
+LIBSSH_API int sftp_seek(sftp_file file, uint32_t new_offset);
 
 /**
  * @brief Seek to a specific location in a file. This is the
@@ -456,7 +530,7 @@ int sftp_seek(SFTP_FILE *file, u32 new_offset);
  *
  * @return             0 on success, < 0 on error.
  */
-int sftp_seek64(SFTP_FILE *file, u64 new_offset);
+LIBSSH_API int sftp_seek64(sftp_file file, uint64_t new_offset);
 
 /**
  * @brief Report current byte position in file.
@@ -467,7 +541,7 @@ int sftp_seek64(SFTP_FILE *file, u64 new_offset);
  *                      of the file associated with the file descriptor. < 0 on
  *                      error.
  */
-unsigned long sftp_tell(SFTP_FILE *file);
+LIBSSH_API unsigned long sftp_tell(sftp_file file);
 
 /**
  * @brief Report current byte position in file.
@@ -478,7 +552,7 @@ unsigned long sftp_tell(SFTP_FILE *file);
  *                      of the file associated with the file descriptor. < 0 on
  *                      error.
  */
-u64 sftp_tell64(SFTP_FILE *file);
+LIBSSH_API uint64_t sftp_tell64(sftp_file file);
 
 /**
  * @brief Rewinds the position of the file pointer to the beginning of the
@@ -486,12 +560,7 @@ u64 sftp_tell64(SFTP_FILE *file);
  *
  * @param file          Open sftp file handle.
  */
-void sftp_rewind(SFTP_FILE *file);
-
-/**
- * @deprecated          Use sftp_unlink() instead.
- */
-int sftp_rm(SFTP_SESSION *sftp, const char *file) SFTP_DEPRECATED;
+LIBSSH_API void sftp_rewind(sftp_file file);
 
 /**
  * @brief Unlink (delete) a file.
@@ -502,7 +571,7 @@ int sftp_rm(SFTP_SESSION *sftp, const char *file) SFTP_DEPRECATED;
  *
  * @return              0 on success, < 0 on error with ssh and sftp error set.
  */
-int sftp_unlink(SFTP_SESSION *sftp, const char *file);
+LIBSSH_API int sftp_unlink(sftp_session sftp, const char *file);
 
 /**
  * @brief Remove a directoy.
@@ -513,7 +582,7 @@ int sftp_unlink(SFTP_SESSION *sftp, const char *file);
  *
  * @return              0 on success, < 0 on error with ssh and sftp error set.
  */
-int sftp_rmdir(SFTP_SESSION *sftp, const char *directory);
+LIBSSH_API int sftp_rmdir(sftp_session sftp, const char *directory);
 
 /**
  * @brief Create a directory.
@@ -528,7 +597,7 @@ int sftp_rmdir(SFTP_SESSION *sftp, const char *directory);
  *
  * @return              0 on success, < 0 on error with ssh and sftp error set.
  */
-int sftp_mkdir(SFTP_SESSION *sftp, const char *directory, mode_t mode);
+LIBSSH_API int sftp_mkdir(sftp_session sftp, const char *directory, mode_t mode);
 
 /**
  * @brief Rename or move a file or directory.
@@ -543,7 +612,7 @@ int sftp_mkdir(SFTP_SESSION *sftp, const char *directory, mode_t mode);
  *
  * @return              0 on success, < 0 on error with ssh and sftp error set.
  */
-int sftp_rename(SFTP_SESSION *sftp, const char *original, const  char *newname);
+LIBSSH_API int sftp_rename(sftp_session sftp, const char *original, const  char *newname);
 
 /**
  * @brief Set file attributes on a file, directory or symbolic link.
@@ -557,7 +626,7 @@ int sftp_rename(SFTP_SESSION *sftp, const char *original, const  char *newname);
  *
  * @return              0 on success, < 0 on error with ssh and sftp error set.
  */
-int sftp_setstat(SFTP_SESSION *sftp, const char *file, SFTP_ATTRIBUTES *attr);
+LIBSSH_API int sftp_setstat(sftp_session sftp, const char *file, sftp_attributes attr);
 
 /**
  * @brief Change the file owner and group
@@ -572,7 +641,7 @@ int sftp_setstat(SFTP_SESSION *sftp, const char *file, SFTP_ATTRIBUTES *attr);
  *
  * @return              0 on success, < 0 on error with ssh and sftp error set.
  */
-int sftp_chown(SFTP_SESSION *sftp, const char *file, uid_t owner, gid_t group);
+LIBSSH_API int sftp_chown(sftp_session sftp, const char *file, uid_t owner, gid_t group);
 
 /**
  * @brief Change permissions of a file
@@ -587,7 +656,7 @@ int sftp_chown(SFTP_SESSION *sftp, const char *file, uid_t owner, gid_t group);
  *
  * @return              0 on success, < 0 on error with ssh and sftp error set.
  */
-int sftp_chmod(SFTP_SESSION *sftp, const char *file, mode_t mode);
+LIBSSH_API int sftp_chmod(sftp_session sftp, const char *file, mode_t mode);
 
 /**
  * @brief Change the last modification and access time of a file.
@@ -601,7 +670,7 @@ int sftp_chmod(SFTP_SESSION *sftp, const char *file, mode_t mode);
  *
  * @return              0 on success, < 0 on error with ssh and sftp error set.
  */
-int sftp_utimes(SFTP_SESSION *sftp, const char *file, const struct timeval *times);
+LIBSSH_API int sftp_utimes(sftp_session sftp, const char *file, const struct timeval *times);
 
 /**
  * @brief Create a symbolic link.
@@ -614,7 +683,7 @@ int sftp_utimes(SFTP_SESSION *sftp, const char *file, const struct timeval *time
  *
  * @return              0 on success, < 0 on error with ssh and sftp error set.
  */
-int sftp_symlink(SFTP_SESSION *sftp, const char *target, const char *dest);
+LIBSSH_API int sftp_symlink(sftp_session sftp, const char *target, const char *dest);
 
 /**
  * @brief Read the value of a symbolic link.
@@ -625,7 +694,34 @@ int sftp_symlink(SFTP_SESSION *sftp, const char *target, const char *dest);
  *
  * @return              The target of the link, NULL on error.
  */
-char *sftp_readlink(SFTP_SESSION *sftp, const char *path);
+LIBSSH_API char *sftp_readlink(sftp_session sftp, const char *path);
+
+/**
+ * @brief Get information about a mounted file system.
+ *
+ * @param  sftp         The sftp session handle.
+ *
+ * @param  path         The pathname of any file within the mounted file system.
+ *
+ * @return A statvfs structure or NULL on error.
+ */
+LIBSSH_API sftp_statvfs_t sftp_statvfs(sftp_session sftp, const char *path);
+
+/**
+ * @brief Get information about a mounted file system.
+ *
+ * @param  file         An opened file.
+ *
+ * @return A statvfs structure or NULL on error.
+ */
+LIBSSH_API sftp_statvfs_t sftp_fstatvfs(sftp_file file);
+
+/**
+ * @brief Free the memory of an allocated statvfs.
+ *
+ * @param  statvfs_o      The statvfs to free.
+ */
+LIBSSH_API void sftp_statvfs_free(sftp_statvfs_t statvfs_o);
 
 /**
  * @brief Canonicalize a sftp path.
@@ -636,7 +732,7 @@ char *sftp_readlink(SFTP_SESSION *sftp, const char *path);
  *
  * @return              The canonicalize path, NULL on error.
  */
-char *sftp_canonicalize_path(SFTP_SESSION *sftp, const char *path);
+LIBSSH_API char *sftp_canonicalize_path(sftp_session sftp, const char *path);
 
 /**
  * @brief Get the version of the SFTP protocol supported by the server
@@ -645,7 +741,7 @@ char *sftp_canonicalize_path(SFTP_SESSION *sftp, const char *path);
  *
  * @return              The server version.
  */
-int sftp_server_version(SFTP_SESSION *sftp);
+LIBSSH_API int sftp_server_version(sftp_session sftp);
 
 #ifdef WITH_SERVER
 /**
@@ -657,7 +753,7 @@ int sftp_server_version(SFTP_SESSION *sftp);
  *
  * @return              A new sftp server session.
  */
-SFTP_SESSION *sftp_server_new(SSH_SESSION *session, CHANNEL *chan);
+LIBSSH_API sftp_session sftp_server_new(ssh_session session, ssh_channel chan);
 
 /**
  * @brief Intialize the sftp server.
@@ -666,32 +762,32 @@ SFTP_SESSION *sftp_server_new(SSH_SESSION *session, CHANNEL *chan);
  *
  * @return             0 on success, < 0 on error.
  */
-int sftp_server_init(SFTP_SESSION *sftp);
+LIBSSH_API int sftp_server_init(sftp_session sftp);
 #endif  /* WITH_SERVER */
 
 /* this is not a public interface */
 #define SFTP_HANDLES 256
-SFTP_PACKET *sftp_packet_read(SFTP_SESSION *sftp);
-int sftp_packet_write(SFTP_SESSION *sftp,u8 type, BUFFER *payload);
-void sftp_packet_free(SFTP_PACKET *packet);
-int buffer_add_attributes(BUFFER *buffer, SFTP_ATTRIBUTES *attr);
-SFTP_ATTRIBUTES *sftp_parse_attr(SFTP_SESSION *session, BUFFER *buf,int expectname);
+sftp_packet sftp_packet_read(sftp_session sftp);
+int sftp_packet_write(sftp_session sftp,uint8_t type, ssh_buffer payload);
+void sftp_packet_free(sftp_packet packet);
+int buffer_add_attributes(ssh_buffer buffer, sftp_attributes attr);
+sftp_attributes sftp_parse_attr(sftp_session session, ssh_buffer buf,int expectname);
 /* sftpserver.c */
 
-SFTP_CLIENT_MESSAGE *sftp_get_client_message(SFTP_SESSION *sftp);
-void sftp_client_message_free(SFTP_CLIENT_MESSAGE *msg);
-int sftp_reply_name(SFTP_CLIENT_MESSAGE *msg, const char *name,
-    SFTP_ATTRIBUTES *attr);
-int sftp_reply_handle(SFTP_CLIENT_MESSAGE *msg, STRING *handle);
-STRING *sftp_handle_alloc(SFTP_SESSION *sftp, void *info);
-int sftp_reply_attr(SFTP_CLIENT_MESSAGE *msg, SFTP_ATTRIBUTES *attr);
-void *sftp_handle(SFTP_SESSION *sftp, STRING *handle);
-int sftp_reply_status(SFTP_CLIENT_MESSAGE *msg, u32 status, const char *message);
-int sftp_reply_names_add(SFTP_CLIENT_MESSAGE *msg, const char *file,
-    const char *longname, SFTP_ATTRIBUTES *attr);
-int sftp_reply_names(SFTP_CLIENT_MESSAGE *msg);
-int sftp_reply_data(SFTP_CLIENT_MESSAGE *msg, const void *data, int len);
-void sftp_handle_remove(SFTP_SESSION *sftp, void *handle);
+sftp_client_message sftp_get_client_message(sftp_session sftp);
+void sftp_client_message_free(sftp_client_message msg);
+int sftp_reply_name(sftp_client_message msg, const char *name,
+    sftp_attributes attr);
+int sftp_reply_handle(sftp_client_message msg, ssh_string handle);
+ssh_string sftp_handle_alloc(sftp_session sftp, void *info);
+int sftp_reply_attr(sftp_client_message msg, sftp_attributes attr);
+void *sftp_handle(sftp_session sftp, ssh_string handle);
+int sftp_reply_status(sftp_client_message msg, uint32_t status, const char *message);
+int sftp_reply_names_add(sftp_client_message msg, const char *file,
+    const char *longname, sftp_attributes attr);
+int sftp_reply_names(sftp_client_message msg);
+int sftp_reply_data(sftp_client_message msg, const void *data, int len);
+void sftp_handle_remove(sftp_session sftp, void *handle);
 
 /* SFTP commands and constants */
 #define SSH_FXP_INIT 1
@@ -797,7 +893,9 @@ void sftp_handle_remove(SFTP_SESSION *sftp, void *handle);
 #define SFTP_READLINK SSH_FXP_READLINK
 #define SFTP_SYMLINK SSH_FXP_SYMLINK
 
-
+/* openssh flags */
+#define SSH_FXE_STATVFS_ST_RDONLY 0x1 /* read-only */
+#define SSH_FXE_STATVFS_ST_NOSUID 0x2 /* no setuid */
 
 #ifdef __cplusplus
 } ;
